@@ -1,4 +1,5 @@
 import re
+from collections import namedtuple
 
 from pypeg2 import List, contiguous, csl, name, optional, parse
 
@@ -27,40 +28,50 @@ class AllFields(str):
     grammar = '*'
 
 
-class BaseArgument(List):
+class ArgumentWithoutQuotes(List):
+    grammar = name(), ':', re.compile(r'true|false|null|[-+]?[0-9]*\.?[0-9]+')
+
+    def number(self, val):
+        try:
+            return int(val)
+        except ValueError:
+            return float(val)
+
     @property
     def value(self):
-        try:
-            return self[0]
-        except IndexError:
-            return ""
+        raw_val = self[0]
+        FIXED_DATA_TYPES = {
+            'true': True,
+            'false': False,
+            'null': None
+        }
+        if raw_val in FIXED_DATA_TYPES:
+            return FIXED_DATA_TYPES[raw_val]
+        return self.number(raw_val)
 
 
-class ArgumentWithoutQuotes(BaseArgument):
-    grammar = name(), ':', re.compile(r'[^,:"\'\)]+')
+class ArgumentWithQuotes(List):
+    grammar = name(), ':', re.compile(r'"([^"\\]|\\.|\\\n)*"|\'([^\'\\]|\\.|\\\n)*\'')
 
-
-class ArgumentWithSingleQuotes(BaseArgument):
-    grammar = name(), ':', "'", optional(re.compile(r'[^\']+')), "'"
-
-
-class ArgumentWithDoubleQuotes(BaseArgument):
-    grammar = name(), ':', '"', optional(re.compile(r'[^"]+')), '"'
+    @property
+    def value(self):
+        # Slicing is for removing quotes
+        # at the begining and end of a string
+        return self[0][1:-1]
 
 
 class Arguments(List):
     grammar = optional(csl(
         [
             ArgumentWithoutQuotes,
-            ArgumentWithSingleQuotes,
-            ArgumentWithDoubleQuotes
+            ArgumentWithQuotes,
         ],
-        separator=','
+        separator=[',', '']
     ))
 
 
 class ArgumentsBlock(List):
-    grammar = optional('(', Arguments, ')')
+    grammar = optional('(', Arguments, optional(','), ')')
 
     @property
     def arguments(self):
@@ -91,12 +102,12 @@ class ParentField(List):
 class BlockBody(List):
     grammar = optional(csl(
         [ParentField, IncludedField, ExcludedField, AllFields],
-        separator=','
+        separator=[',', '']
     ))
 
 
 class Block(List):
-    grammar = ArgumentsBlock, '{', BlockBody, '}'
+    grammar = ArgumentsBlock, '{', BlockBody, optional(','), '}'
 
     @property
     def arguments(self):
@@ -116,48 +127,55 @@ class Block(List):
 ParentField.grammar = IncludedField, Block
 
 
+Query = namedtuple(
+    "Query",
+    ("field_name", "included_fields", "excluded_fields", "aliases", "arguments")
+)
+
+
 class QueryParser(object):
     def parse(self, query):
         parse_tree = parse(query, Block)
-        return self._transform_block(parse_tree)
+        return self._transform_block(parse_tree, parent_field=None)
 
-    def _transform_block(self, block):
-        fields = {
-            "include": [],
-            "exclude": [],
-            "arguments": {},
-            "aliases": {},
-        }
+    def _transform_block(self, block, parent_field=None):
+        query = Query(
+            field_name=parent_field,
+            included_fields=[],
+            excluded_fields=[],
+            aliases={},
+            arguments={}
+        )
 
         for argument in block.arguments:
             argument = {str(argument.name): argument.value}
-            fields['arguments'].update(argument)
+            query.arguments.update(argument)
 
         for field in block.body:
             # A field may be a parent or included field or excluded field
             if isinstance(field, (ParentField, IncludedField)):
                 # Find all aliases
                 if field.alias:
-                    fields["aliases"].update({str(field.name): str(field.alias)})
+                    query.aliases.update({str(field.name): str(field.alias)})
 
             field = self._transform_field(field)
 
-            if isinstance(field, dict):
+            if isinstance(field, Query):
                 # A field is a parent
-                fields["include"].append(field)
+                query.included_fields.append(field)
             elif isinstance(field, IncludedField):
-                fields["include"].append(str(field.name))
+                query.included_fields.append(str(field.name))
             elif isinstance(field, ExcludedField):
-                fields["exclude"].append(str(field.name))
+                query.excluded_fields.append(str(field.name))
             elif isinstance(field, AllFields):
                 # include all fields
-                fields["include"].append("*")
+                query.included_fields.append("*")
 
-        if fields["exclude"] and "*" not in fields["include"]:
-            fields["include"].append("*")
+        if query.excluded_fields and "*" not in query.included_fields:
+            query.included_fields.append("*")
 
-        field_names = set(fields["aliases"].values())
-        field_aliases = set(fields["aliases"].keys())
+        field_names = set(query.aliases.values())
+        field_aliases = set(query.aliases.keys())
         faulty_fields = field_names.intersection(field_aliases)
         if faulty_fields:
             # We check this here because if we let it pass during
@@ -170,7 +188,7 @@ class QueryParser(object):
                 "The list of fields which led to this error is %s."
             ) % str(list(faulty_fields))
             raise QueryFormatError(msg)
-        return fields
+        return query
 
     def _transform_field(self, field):
         # A field may be a parent or included field or excluded field
@@ -179,6 +197,7 @@ class QueryParser(object):
         return field
 
     def _transform_parent_field(self, parent_field):
-        parent_field_name = str(parent_field.name)
-        parent_field_value = self._transform_block(parent_field.block)
-        return {parent_field_name: parent_field_value}
+        return self._transform_block(
+            parent_field.block,
+            parent_field=str(parent_field.name)
+        )
